package ch.epfl.cs108;

import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayDeque;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Deque;
import java.util.List;

public final class LZWInputStream extends InputStream {
    private final static int DICT_CAPACITY = 4096;
    private final Bits12InputStream in;
    private final List<List<Integer>> dict;
    private final Deque<Integer> currList;
    private List<Integer> prevList;

    public LZWInputStream(InputStream in) {
        List<List<Integer>> dict = new ArrayList<>(DICT_CAPACITY);
        for (int i = 0; i <= 0xFF; ++i)
            dict.add(Collections.singletonList(i));

        this.in = new Bits12InputStream(in);
        this.dict = dict;
        this.currList = new ArrayDeque<>();
        this.prevList = null;
    }

    @Override
    public int read() throws IOException {
        if (currList.isEmpty()) {
            int i = in.readU12();
            if (i == -1)
                return -1;

            List<Integer> newList = i < dict.size()
                    ? dict.get(i) : copyAndAppendHead(prevList, prevList);

            if (dict.size() < DICT_CAPACITY && prevList != null)
                dict.add(copyAndAppendHead(prevList, newList));

            currList.addAll(newList);
            prevList = newList;
        }
        return currList.removeFirst();
    }

    private <T> List<T> copyAndAppendHead(List<T> l, List<T> l2) {
        List<T> r = new ArrayList<>(l.size() + 1);
        r.addAll(l);
        r.add(l2.get(0));
        return Collections.unmodifiableList(r);
    }

    @Override
    public void close() throws IOException {
        in.close();
    }
}
